/******************************************************************************
 *                                                                            *
 *  MMX and SSE Courseware                                                    *
 *                                                                            *
 * Course: Advanced Computer Architectures (36SPA, X36SPA, XD36SPA, XE36SPA)  *
 *                                                                            *
 * Department of Computer Science and Engineering                             *
 * Faculty of Electrical Engineering                                          *
 * Czech Technical University in Prague                                       * 
 *                                                                            *
 * Miroslav Skrbek (C)2005,2007                                               *  
 *                                                                            *
 ******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <assert.h>
#include <string.h>

#include <gtk/gtk.h>

#include "utils.h"
#include "main.h"

#define TITLE_STR_LEN 32
 
typedef int COLORREF;

int gr_mode;
int windows_opened = 0;

//// Window Handlers ////////////////////////////////////////////////////////////////////////////////////

static gboolean delete_event( GtkWidget *widget,
                              GdkEvent  *event,
                              gpointer   data )
{
    /* emit detroy event */
    return FALSE;
}


static void destroy( GtkWidget *widget,
                     gpointer   data )
{
    windows_opened--;
    if (windows_opened <= 0) {		
    	gtk_main_quit();
    }	
}

//// Image Functions /////////////////////////////////////////////////////////////////////////////////////

/* This function loads the RGB image from a file (*.bmp) */

void load_rgb_image(char* filename, int* width, int* height, COLORREF* image) {
 
	leave_mmx_mode();

	if (width == NULL || height == NULL) {
		printf("Error: width and height parameter of the %s function must be a valid pointer\n", __FUNCTION__);
		exit(1);
	}	

	printf("Loading RGB image (%s)\n", filename);
 
	GtkWidget *gtk_image = gtk_image_new_from_file(filename);
 
    	GdkPixbuf *pixbuf = gtk_image_get_pixbuf(GTK_IMAGE(gtk_image));
 
	int w  = gdk_pixbuf_get_width(pixbuf);
	int h = gdk_pixbuf_get_height(pixbuf);
	int depth = gdk_pixbuf_get_bits_per_sample(pixbuf);
	int channels = gdk_pixbuf_get_n_channels(pixbuf);

	printf("Image info: %dx%d, depth: %d, channels: %d\n", w, h, depth, channels);
	if (depth != 8 || channels != 3) {
		printf("Erorr: Truecolor RGB image expected.\n");
		exit(1); 

	}

	*width = w;
	*height = h;

	if (image == NULL) {
		return;
	}

	int x, y;
 	guchar *pixels = gdk_pixbuf_get_pixels(pixbuf);

	for(y = 0; y < h; y++) {
		for(x = 0; x < w; x++) {
			int index = y * w + x;
			int r = pixels[index * 3 + 0];
			int g = pixels[index * 3 + 1];
			int b = pixels[index * 3 + 2];
			int rgb = r << 16 | g << 8 | b;
			image[index] = rgb;
		}
	}
					 
	enter_mmx_mode();
}	

///////////////////////////////////////////////////////////////////////////////////////////

/* This function loads the RGB from a file and converts it to the
   grayscale image. Each 8-bit pixel represents 256-gray levels.
*/ 

void load_gray_image(char* filename, int* width, int* height, unsigned char* image) {

	leave_mmx_mode();

	printf("Loading RGB image (%s)\n", filename);

	if (width == NULL || height == NULL) {
		printf("Error: width and height parameter of the %s function must be a valid pointer\n", __FUNCTION__);
		exit(1);
	}	

	printf("Loading RGB image (%s)\n", filename);
 
	GtkWidget *gtk_image = gtk_image_new_from_file(filename);
 
    	GdkPixbuf *pixbuf = gtk_image_get_pixbuf(GTK_IMAGE(gtk_image));
 
	int w = gdk_pixbuf_get_width(pixbuf);
	int h = gdk_pixbuf_get_height(pixbuf);
	int depth = gdk_pixbuf_get_bits_per_sample(pixbuf);
	int channels = gdk_pixbuf_get_n_channels(pixbuf);

	printf("Image info: %d, %d, depth: %d, channels: %d\n", w, h, depth, channels);
	if (depth != 8 || channels != 3) {
		printf("Error: truecolor RGB image expected.\n");
		exit(1); 

	}

	*width = w;
	*height = h;

	if (image == NULL) {
		return;
	}

	printf("Converting to grayscale ... ");
	int x, y;
 	guchar *pixels = gdk_pixbuf_get_pixels(pixbuf);

	for(y = 0; y < h; y++) {
		for(x = 0; x < w; x++) {
			int index = y * w + x;
			int r = pixels[index * 3 + 0];
			int g = pixels[index * 3 + 1];
			int b = pixels[index * 3 + 2];
			image[index] = 0.3*r + 0.59*g + 0.11*b;
		}
	}
		
	printf("done.\n");
					 
	enter_mmx_mode();
}	

static void free_pixbuf(guchar *pixels, gpointer data) {
	free(pixels);
	printf("freeing pixbuf\n");	
}

//////////////////////////////////////////////////////////////////////////////////////////////

void show_rgb_image(int num, int width, int height, COLORREF* image) {

	leave_mmx_mode();

	if (!gr_mode) {
		printf("\nError: application is not in graphic mode\n\n");
		printHelp();
		exit(1);
	}	 

 	if (image == NULL) {
		printf("Non-null image required\n");
		exit(1);
	}

	int x, y;

    	GtkWidget* window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

	char title[TITLE_STR_LEN + 1];
	snprintf(title, TITLE_STR_LEN, "Image %d (RGB)", num);
 	gtk_window_set_title(GTK_WINDOW(window), title);

   	g_signal_connect(G_OBJECT (window), "destroy", G_CALLBACK (destroy), NULL);
    	g_signal_connect (G_OBJECT (window), "delete_event", G_CALLBACK (delete_event), NULL);

	guchar* image_data = malloc(3 * width * height);
	
	for(y = 0; y < height; y++) {
		for(x = 0; x < width; x++) {
			guint c = image[y * width + x];
			guint i = 3 * (y * width + x);
			image_data[i + 0] = (c >> 16) & 0xFF; 
			image_data[i + 1] = (c >> 8) & 0xFF; 
			image_data[i + 2] = (c) & 0xFF; 
		}
	}
	
 
	GdkPixbuf *pixbuf = gdk_pixbuf_new_from_data(image_data, GDK_COLORSPACE_RGB,
				0, 8, width, height, 3 * width, free_pixbuf, NULL);

	
	GtkWidget *gtk_image = gtk_image_new_from_pixbuf(pixbuf);
	
    	gtk_container_add(GTK_CONTAINER(window), gtk_image);	

    	gtk_widget_show (gtk_image);
   	gtk_widget_show (window);				 
	windows_opened++;

	
	enter_mmx_mode();	
}

///////////////////////////////////////////////////////////////////////////////////////////

/* This function shows a grayscale image. Each 8-bit pixel represents
   256 gray levels */ 
void show_gray_image(int num, int width, int height, unsigned char* image) {

	leave_mmx_mode();

	if (!gr_mode) {
		printf("\nError: application is not in graphic mode\n\n");
		printHelp();
		exit(1);
	}	 

 	if (image == NULL) {
		printf("Non-null image required\n");
		exit(1);
	}

	int x, y;

    	GtkWidget* window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

	char title[TITLE_STR_LEN + 1];
	snprintf(title, TITLE_STR_LEN, "Image %d (Gray)", num);
 	gtk_window_set_title(GTK_WINDOW(window), title);

   	g_signal_connect(G_OBJECT (window), "destroy", G_CALLBACK (destroy), NULL);
    	g_signal_connect (G_OBJECT (window), "delete_event", G_CALLBACK (delete_event), NULL);

	guchar* image_data = malloc(3 * width * height);
	
	for(y = 0; y < height; y++) {
		for(x = 0; x < width; x++) {
			guint c = image[y * width + x];
			guint i = 3 * (y * width + x);
			image_data[i + 0] = c & 0xFF; 
			image_data[i + 1] = c & 0xFF; 
			image_data[i + 2] = c & 0xFF; 
		}
	}
	
 
	GdkPixbuf *pixbuf = gdk_pixbuf_new_from_data(image_data, GDK_COLORSPACE_RGB,
				0, 8, width, height, 3 * width, free_pixbuf, NULL);

	
	GtkWidget *gtk_image = gtk_image_new_from_pixbuf(pixbuf);

    	gtk_container_add(GTK_CONTAINER(window), gtk_image);	

    	gtk_widget_show (gtk_image);
   	gtk_widget_show (window); 
	windows_opened++;

	enter_mmx_mode();	
}

//// MAIN function //////////////////////////////////////////////////////////////////

/* GTK initialization and the main loop */

int main(int argc, char* argv[]) {
	int retval;

	gtk_init(&argc, &argv);
	
	retval = main_body(argc, argv);
	if (retval) {
		return retval; 
	} 
	
	if (gr_mode && windows_opened > 0) {
		gtk_main();
	}
}

/////////////////////////////////////////////////////////////////////////////
