/******************************************************************************
 *                                                                            *
 *  MMX and SSE Courseware                                                    *
 *                                                                            *
 * Course: Advanced Computer Architectures (36SPA, X36SPA, XD36SPA, XE36SPA)  *
 *                                                                            *
 * Department of Computer Science and Engineering                             *
 * Faculty of Electrical Engineering                                          *
 * Czech Technical University in Prague                                       * 
 *                                                                            *
 * Miroslav Skrbek (C)2005                                                    *  
 *                                                                            *
 ******************************************************************************/

/* Important note. This file looks like the main source file, which is typical
   for many C programs, and which usually contains the main() C function. 
   The major difference of this main.c file is that it contains the main_body()
   function instead. The main_body function contains all functionality typical
   for the main() function (parsing programs arguments for example), but 
   the main() function is located in a corresponding user interface source file 
   (gtkui.c or winui.c) depending on platform selected.
*/    

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <assert.h>
#include <string.h>

#include "student.h"
#include "utils.h"
#include "ui.h"

// Modes controlled by command line options
int cpuinfo_mode = 0;	

/////////////////////////////////////////////////////////////////////////////

/* This function explains command line options */
void printHelp() {
   	printf("Usage: main <option1> [<option2>]\n");
	printf("  option1: -gr  ... graphic mode\n");
	printf("           -cl  ... command line mode\n");
	printf("           -cpuinfo  ... CPU Inforamtion\n");
	printf("  option2: -x   ... hexadecimal output (default)\n");
	printf("           -d   ... decimal output\n");
}

/////////////////////////////////////////////////////////////////////////////

/* Floating point exception handler */ 
static void fp_exception(int v) {
  printf("\n*** FP exception, program exited ***\n\n");
  exit(1);
}    

/////////////////////////////////////////////////////////////////////////////

int main_body(int argc, char* argv[]) {

	// Install FP-exception handler
	signal(SIGFPE, fp_exception);
            
   	if (argc < 2) {
        	printHelp();
		return 1;
	}		
	
	if (strcmp(argv[1], "-gr") == 0) {
		gr_mode = 1;
	}
	else 
	if (strcmp(argv[1], "-cl") == 0) {
		gr_mode = 0;
	}
	else 
	if (strcmp(argv[1], "-cpuinfo") == 0) {
		cpuinfo_mode = 1;
	}
	else {
		printf("\nError: invalid option\n\n");
		printHelp();
		return 1;
	}

	/* Set default output format */
	output_format = OUTPUT_FORMAT_HEX;

	if (argc > 2) {
		if (strcmp(argv[2], "-x") == 0) {
			output_format = OUTPUT_FORMAT_HEX;
		}
		else  
		if (strcmp(argv[2], "-u") == 0) {
			output_format = OUTPUT_FORMAT_UDC;
		}
		else
		if (strcmp(argv[2], "-d") == 0) {
			output_format = OUTPUT_FORMAT_DEC;
		}
		else {
			printf("\nError: invalid option\n\n");
			printHelp();
			return 1;
		}
	} 
		
	cpuinfo();

	if (cpuinfo_mode) {
		print_cpuinfo();
		return 0;
	}			

        if (!cpuid_fpu_present || 
	    !cpuid_fsfr_present ||
	    !cpuid_mmx_present) {
	    printf("CPU has not required capabilities (FPU, MMX, and FSAVE/FRSTORE)\n"); 
	    printf("use -cpuinfo option for more details\n");	
	    return 1;
	}
	 	
  	init_mmx_mode(); 
  	enter_mmx_mode();
  	student();
  	leave_mmx_mode();
 
  	return 0;
}

/////////////////////////////////////////////////////////////////////////////
