/******************************************************************************
 *                                                                            *
 *  MMX and SSE Courseware                                                    *
 *                                                                            *
 * Course: Advanced Computer Architectures (36SPA, X36SPA, XD36SPA, XE36SPA)  *
 *                                                                            *
 * Department of Computer Science and Engineering                             *
 * Faculty of Electrical Engineering                                          *
 * Czech Technical University in Prague                                       * 
 *                                                                            *
 * Miroslav Skrbek (C)2005                                                    *  
 *                                                                            *
 ******************************************************************************/
 
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <assert.h>
#include <string.h>

#include "utils.h"

/* Output format */

int output_format = OUTPUT_FORMAT_HEX;

/* CPU features from CPUID */

int cpuid_fpu_present     = 0;
int cpuid_fsfr_present    = 0;
int cpuid_mmx_present     = 0;
int cpuid_mmxex_present   = 0;
int cpuid_sse_present     = 0;
int cpuid_sse2_present    = 0;
int cpuid_sse3_present    = 0;
int cpuid_3dnow_present   = 0;
int cpuid_3dnowex_present = 0;

char cpuid_vendor[13];

/* This data type represents an extended 80-bit floating point number */     
typedef unsigned char extended[10];

/* This structure represents the state of the FPU unit */  
typedef struct  __attribute__ ((packed)) _fp_state {
  unsigned int control_word;  // Ridici slovo
  unsigned int status_word;   // Stavove slovo
  unsigned int tag_word;      // Registr atributu platnosti obsahu registru zasobniku st(x)
  unsigned int fipo;          // FPU Instruction Pointer Offset
  unsigned int fips;          // FPU Instruction Pointer Selector
  unsigned int fopo;          // FPU Operand Pointer Offset
  unsigned int fops;          // FPU Operand Pointer Selector  
  extended st[8];
} fp_state;
 
/* This type represents two possible program context states: 

	FP   - floating point mode (the FPU registers are used as a stack) 
	MMX  - MMX m ode (the FPU registers are used as MMX registers)
	
*/      	
typedef enum { FP, MMX } mode;

/* Messages */
char* msg1 = "The %s function (from %s) can be used in mmx mode only.\n"; 

/* MMX data types */
typedef unsigned char  	MMX_PB[8];
typedef unsigned short 	MMX_PW[4];
typedef unsigned int   	MMX_PD[2];
typedef          float  MMX_FP[2];         

/* SSE data types */
typedef float XMM_PSFP[4]; 

///////////////////////////////////////////////////////////////////////////////

/* FPU state; it is used for FP and MMX context switching */
struct _context {
	mode state;
	fp_state mmx;
	fp_state fp;
} context;

///////////////////////////////////////////////////////////////////////////////

/* Initialize FPU context for both FP and MMX mode */
void init_mmx_mode() {
	asm("fsave %0" : : "m" (context.fp));		
	asm("fsave %0" : : "m" (context.mmx));
	context.state = FP;			
}

/* Enter the MMX mode. This function saves the current FP context
   to the context structure and restores the MMX state from 
   the context structure.
*/   
void enter_mmx_mode() {
	switch (context.state) {
		case FP: 
			asm("fsave %0" : : "m" (context.fp));		
			asm("frstor %0" : : "m" (context.mmx));		
			context.state = MMX;			
			break;				
		case MMX:
			break;
		default: assert(0);	
	}							
}

/* Leave the MMX mode. This function saves the current MMX state to 
   the context structure and restores the FP state from 
   the context structure.
*/      
void leave_mmx_mode() {
	switch (context.state) {
		case FP: 
			break;
		case MMX:
			asm("fsave %0" : : "m" (context.mmx));		
			asm("frstor %0" : : "m" (context.fp));		
			context.state = FP;			
			break;
		default: assert(0);	
	}							
};

///////////////////////////////////////////////////////////////////////////////

/* Macros for print function definitions */

#define DECLARE_PRINT_MM(REG, FORMAT) void print_mm ## REG ## _ ## FORMAT()\
  { asm("\tpushfl\n\tpushal\n\tsubl $4, %esp"); print_mm_ ## FORMAT (REG); asm("\taddl $4, %esp\n\tpopal\n\tpopfl\n");}

#define DECLARE_PRINT_XMM(REG) void print_xmm ## REG()\
  { asm("\tpushfl\n\tpushal\n\tsubl $4, %esp"); print_xmm(REG); asm("\taddl $4, %esp\n\tpopal\n\tpopfl\n");}
  

///////////////////////////////////////////////////////////////////////////////

int cpuinfo() {
	
	unsigned int id_eax;
	unsigned int id_ecx;
	unsigned int id_edx;
	
	char id_vendor[13];	

	memset(id_vendor, 0, 13); 	
	asm("mov $0x00000000, %eax");
	asm("cpuid");
	asm("mov %%ebx,  %0" : "=m" (id_vendor));
	asm("mov %%edx, 4%0" : "=m" (id_vendor));
	asm("mov %%ecx, 8%0" : "=m" (id_vendor));
	memcpy(cpuid_vendor, id_vendor, 13);

	cpuid_fpu_present     = 0;
	cpuid_fsfr_present    = 0;
	cpuid_mmx_present     = 0;
	cpuid_mmxex_present   = 0;
	cpuid_sse_present     = 0;
	cpuid_sse2_present    = 0;
	cpuid_sse3_present    = 0;
	cpuid_3dnow_present   = 0;
	cpuid_3dnowex_present = 0;
	
	if (strcmp(cpuid_vendor, "AuthenticAMD") == 0) {

		asm("mov $0x00000000, %eax");
		asm("cpuid");
		asm("mov %%eax,  %0" : "=m" (id_eax));
	
		if (id_eax < 1) {
			printf("Error: insufficient number of CPUID functions\n");
			exit(1);
		}		

		asm("mov $0x00000001, %eax");
		asm("cpuid");
		asm("mov %%ecx, %0" : "=m" (id_ecx));
		asm("mov %%edx, %0" : "=m" (id_edx));

		cpuid_fpu_present  = ((id_edx >>  0) & 0x1);
		cpuid_mmx_present  = ((id_edx >> 23) & 0x1);
		cpuid_fsfr_present = ((id_edx >> 24) & 0x1);
		cpuid_sse_present = ((id_edx >> 25) & 0x1);
		cpuid_sse2_present = ((id_edx >> 26) & 0x1);
		cpuid_sse3_present = ((id_ecx >>  0) & 0x1);
		
		asm("mov $0x80000000, %eax");
		asm("cpuid");
		asm("mov %%eax,  %0" : "=m" (id_eax));
		
		if (id_eax < 0x80000002) {
			printf("Error: insufficient number of extended CPUID functions\n");
			exit(1);
		}	
		asm("mov $0x80000001, %eax");
		asm("cpuid");
		asm("mov %%ecx, %0" : "=m" (id_ecx));
		asm("mov %%edx, %0" : "=m" (id_edx));

		cpuid_mmxex_present = ((id_edx >> 22) & 0x1);
		cpuid_3dnow_present = ((id_edx >> 31) & 0x1);
		cpuid_3dnowex_present = ((id_edx >> 30) & 0x1);

	}
	else
	if (strcmp(cpuid_vendor, "GenuineIntel") == 0) {
	
		asm("mov $0x00000000, %eax");
		asm("cpuid");
		asm("mov %%eax,  %0" : "=m" (id_eax));
	
		if (id_eax < 1) {
			printf("Error: insufficient number of CPUID functions\n");
			exit(1);
		}		

		asm("mov $0x00000001, %eax");
		asm("cpuid");
		asm("mov %%ecx, %0" : "=m" (id_ecx));
		asm("mov %%edx, %0" : "=m" (id_edx));

		cpuid_fpu_present = ((id_edx >>  0) & 0x1);
		cpuid_mmx_present = ((id_edx >> 23) & 0x1);
		cpuid_fsfr_present = ((id_edx >> 24) & 0x1);
		cpuid_sse_present = ((id_edx >> 25) & 0x1);
		cpuid_sse2_present = ((id_edx >> 26) & 0x1);
		cpuid_sse3_present = ((id_ecx >>  0) & 0x1);
		
	}
	else {
		printf("Error: unknown processor vendor\n");
		exit(1);
	}	
			
};

///////////////////////////////////////////////////////////////////////////////

void print_cpuinfo() {
	printf("\n");
	printf(" + CPUINFO ---------------------------+\n");
	printf(" | CPU Vendor       : %-15s |\n", cpuid_vendor);
	printf(" | FPU              : %-15s |\n", (cpuid_fpu_present)     ? "present" : "not present");
	printf(" | FSAVE/FRSTOR     : %-15s |\n", (cpuid_fsfr_present)    ? "present" : "not present");
	printf(" | MMX              : %-15s |\n", (cpuid_mmx_present)     ? "present" : "not present");
	printf(" | MMX ext.(AMD)    : %-15s |\n", (cpuid_mmxex_present)   ? "present" : "not present");
	printf(" | SSE              : %-15s |\n", (cpuid_sse_present)     ? "present" : "not present");
	printf(" | SSE2             : %-15s |\n", (cpuid_sse2_present)    ? "present" : "not present");
	printf(" | SSE3             : %-15s |\n", (cpuid_sse3_present)    ? "present" : "not present");
	printf(" | 3DNow!           : %-15s |\n", (cpuid_3dnow_present)   ? "present" : "not present");
	printf(" | 3DNow! ext.(AMD) : %-15s |\n", (cpuid_3dnowex_present) ? "present" : "not present");
	printf(" +------------------------------------+\n");
	printf("\n");
}

///////////////////////////////////////////////////////////////////////////////
	
/* This function loads a requsted MMX register to the mm variable */
static void load_mm_register(int reg, void* mm) {

	switch (reg) {
		case 0: asm("movq %%mm0, %0\n": "=m" (*((MMX_PB*)mm))); break;
		case 1: asm("movq %%mm1, %0\n": "=m" (*((MMX_PB*)mm))); break;
		case 2: asm("movq %%mm2, %0\n": "=m" (*((MMX_PB*)mm))); break;
		case 3: asm("movq %%mm3, %0\n": "=m" (*((MMX_PB*)mm))); break;
		case 4: asm("movq %%mm4, %0\n": "=m" (*((MMX_PB*)mm))); break;
		case 5: asm("movq %%mm5, %0\n": "=m" (*((MMX_PB*)mm))); break;
		case 6: asm("movq %%mm6, %0\n": "=m" (*((MMX_PB*)mm))); break;
		case 7: asm("movq %%mm7, %0\n": "=m" (*((MMX_PB*)mm))); break;
		default: {
			printf("Error: invalid register number %d\n", reg);
			assert(0);
		}	
	}	
}

///////////////////////////////////////////////////////////////////////////////

/* This function loads a requsted XMM register to the xmm variable */
static void load_xmm_register(int reg, void* xmm) {

	switch (reg) {
		case 0: asm("movups %%xmm0, %0\n": "=m" (*((XMM_PSFP*)xmm))); break;
		case 1: asm("movups %%xmm1, %0\n": "=m" (*((XMM_PSFP*)xmm))); break;
		case 2: asm("movups %%xmm2, %0\n": "=m" (*((XMM_PSFP*)xmm))); break;
		case 3: asm("movups %%xmm3, %0\n": "=m" (*((XMM_PSFP*)xmm))); break;
		case 4: asm("movups %%xmm4, %0\n": "=m" (*((XMM_PSFP*)xmm))); break;
		case 5: asm("movups %%xmm5, %0\n": "=m" (*((XMM_PSFP*)xmm))); break;
		case 6: asm("movups %%xmm6, %0\n": "=m" (*((XMM_PSFP*)xmm))); break;
		case 7: asm("movups %%xmm7, %0\n": "=m" (*((XMM_PSFP*)xmm))); break;
		default: {
			printf("Error: invalid register number %d\n", reg);
			assert(0);
		}			
	}	
}

///////////////////////////////////////////////////////////////////////////////

/* This function prints a requested MMX register to stdout in the packed-byte format */
void print_mm_pb(int reg) {
	MMX_PB mmx;

	if (context.state != MMX) {
		printf(msg1, __FUNCTION__, __FILE__); 
		exit(1);
	}
	
	load_mm_register(reg, &mmx);
		
	leave_mmx_mode(); 
	switch (output_format) {
		case OUTPUT_FORMAT_DEC: 
			printf("| MM%d: %4d %4d %4d %4d %4d %4d %4d %4d |\n", reg, 
	  			(char)mmx[7], (char)mmx[6], (char)mmx[5], (char)mmx[4],
	  			(char)mmx[3], (char)mmx[2], (char)mmx[1], (char)mmx[0]);
			break;
		case OUTPUT_FORMAT_UDC: 
			printf("| MM%d:  %3u  %3u  %3u  %3u  %3u  %3u  %3u  %3u |\n", reg, 
	  			mmx[7], mmx[6], mmx[5], mmx[4],
	  			mmx[3], mmx[2], mmx[1], mmx[0]);
			break;
		case OUTPUT_FORMAT_HEX: 
			printf("| MM%d:   %02X   %02X   %02X   %02X   %02X   %02X   %02X   %02X |\n", reg, 
	  			mmx[7], mmx[6], mmx[5], mmx[4],
	  			mmx[3], mmx[2], mmx[1], mmx[0]);
			break;
	}
	enter_mmx_mode();  
}

/* This function prints a requested MMX register to stdout in the packed word format */
void print_mm_pw(int reg) {
	MMX_PW mmx;
	if (context.state != MMX) {
		printf(msg1, __FUNCTION__, __FILE__); 
		exit(1);
	}

	load_mm_register(reg, &mmx);
		
	leave_mmx_mode();
	switch (output_format) {
		case OUTPUT_FORMAT_DEC: 
			printf("| MM%d: %6d %6d %6d %6d |\n", reg, 
	  			(short)mmx[3], (short)mmx[2], (short)mmx[1], (short)mmx[0]);
			break;
		case OUTPUT_FORMAT_UDC: 
			printf("| MM%d:  %5u  %5u  %5u  %5u |\n", reg, 
	  			mmx[3], mmx[2], mmx[1], mmx[0]);
			break;
		case OUTPUT_FORMAT_HEX: 
			printf("| MM%d:   %04X   %04X   %04X   %04X |\n", reg, 
	  			mmx[3], mmx[2], mmx[1], mmx[0]);
			break;
	}
	enter_mmx_mode();  
}

/* This function prints a requested MMX register to stdout in the packed dword format */
void print_mm_pd(int reg) {
	MMX_PD mmx;
	if (context.state != MMX) {
		printf(msg1, __FUNCTION__, __FILE__); 
		exit(1);
	}

	load_mm_register(reg, &mmx);
		
	leave_mmx_mode();
	switch (output_format) {
		case OUTPUT_FORMAT_DEC: 
			printf("| MM%d: %11d %11d |\n", reg, 
	  			mmx[1], mmx[0]);
			break;
		case OUTPUT_FORMAT_UDC: 
			printf("| MM%d:  %10u  %10u |\n", reg, 
	  			mmx[1], mmx[0]);
			break;
		case OUTPUT_FORMAT_HEX: 
			printf("| MM%d:    %08X    %08X |\n", reg, 
	  			mmx[1], mmx[0]);
			break;
	}
	enter_mmx_mode();  
}

/* This function prints a requested MMX register to stdout in the packed qword format */
void print_mm_pq(int reg) {
	MMX_PD mmx;
	if (context.state != MMX) {
		printf(msg1, __FUNCTION__, __FILE__); 
		exit(1);
	}

	load_mm_register(reg, &mmx);
		
	leave_mmx_mode();
	switch (output_format) {
		case OUTPUT_FORMAT_DEC: {
				unsigned long long v0 = mmx[0];
				unsigned long long v1 = mmx[1];
				unsigned long long v = (v1 << 32) | v0;
				printf("| MM%d: %21lld |\n", reg, v);
			}
			break;
		case OUTPUT_FORMAT_UDC: {
				unsigned long long v0 = mmx[0];
				unsigned long long v1 = mmx[1];
				unsigned long long v = (v1 << 32) | v0;
				printf("| MM%d:  %20llu |\n", reg, v);
			}
			break;
		case OUTPUT_FORMAT_HEX: 
			printf("| MM%d:      %08X%08X |\n", reg, 
	  			mmx[1], mmx[0]);
			break;
	}
	enter_mmx_mode();  
}

/* This function prints a requested MMX register to stdout in the single FP format */
void print_mm_fp(int reg) {
	MMX_FP mmx;
	if (context.state != MMX) {
		printf(msg1, __FUNCTION__, __FILE__); 
		exit(1);
	}

	load_mm_register(reg, &mmx);
		
	leave_mmx_mode();
	printf("| MM%d: %15.6f  %15.6f |\n", reg, 
	  mmx[1], mmx[0]);
	enter_mmx_mode();  
}

///////////////////////////////////////////////////////////////////////////////

/* This function prints a requested XMM register to stdout 
  in the single-precision FP format */
void print_xmm(int reg) {

	XMM_PSFP xmm;

	if (context.state != MMX) {
		printf(msg1, __FUNCTION__, __FILE__); 
		exit(1);
	}
	
	if (!cpuid_sse_present) {
		printf("Error: SSE is not supported by CPU\n");
		printf("Use -cpuinfo option to see CPU features\n");
		exit(1);
	}		

	load_xmm_register(reg, &xmm);
	
	leave_mmx_mode();
	printf("| XMM%d: | %14.6e | %14.6e | %14.6e | %14.6e |\n", reg, 
		xmm[3], xmm[2], xmm[1], xmm[0]);
	enter_mmx_mode();
	
}


///////////////////////////////////////////////////////////////////////////////

/* This function prints all XMM registers in the single-precision FP format */
void cprint_xmm_regs() {
	int i;
	if (!cpuid_sse_present) {
		printf("Error: SSE is not supported by CPU\n");
		printf("Use -cpuinfo option to see CPU features\n");
		exit(1);
	}		

	printf("+---------------------------------------------------------------------------+\n");
	for(i = 0; i < 8; i++) print_xmm(i);
	printf("+---------------------------------------------------------------------------+\n");
}	

/* This function prints all MMX registers in the packet byte format */
void cprint_mm_regs_pb() {
	int i;
	printf("+----------------------------------------------+\n");
	for(i = 0; i < 8; i++) print_mm_pb(i);
	printf("+----------------------------------------------+\n");
}	

/* This function prints all MMX registers in the packet word format */
void cprint_mm_regs_pw() {
	int i;
	printf("+----------------------------------+\n");
	for(i = 0; i < 8; i++) print_mm_pw(i);
	printf("+----------------------------------+\n");
}	
 
/* This function prints all MMX registers in the packet dword format */
void cprint_mm_regs_pd() {
	int i;
	asm("\tpushfl\n\tpushal");
	printf("+------------------------------+\n");
	for(i = 0; i < 8; i++) print_mm_pd(i);
	printf("+------------------------------+\n");
	asm("\tpopal\n\tpopfl");
}	
 
/* This function prints all MMX registers in the packet qword format */
void cprint_mm_regs_pq() {
	int i;
	asm("\tpushfl\n\tpushal");
	printf("+----------------------------+\n");
	for(i = 0; i < 8; i++) print_mm_pq(i);
	printf("+----------------------------+\n");
	asm("\tpopal\n\tpopfl");
}	

/* This function prints all MMX registers in the FP format */
void cprint_mm_regs_fp() {
	int i;
	asm("\tpushfl\n\tpushal");
	printf("+-----------------------------------------+\n");
	for(i = 0; i < 8; i++) print_mm_fp(i);
	printf("+-----------------------------------------+\n");
	asm("\tpopal\n\tpopfl");
}	

///////////////////////////////////////////////////////////////////////////////
