/******************************************************************************
 *                                                                            *
 *  MMX and SSE Courseware                                                    *
 *                                                                            *
 * Course: Advanced Computer Architectures (36SPA, X36SPA, XD36SPA, XE36SPA)  *
 *                                                                            *
 * Department of Computer Science and Engineering                             *
 * Faculty of Electrical Engineering                                          *
 * Czech Technical University in Prague                                       * 
 *                                                                            *
 * Miroslav Skrbek (C)2005, 2007                                              *  
 *                                                                            *
 ******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <assert.h>
#include <string.h>
#include <windows.h>
#include <wingdi.h>

#include "utils.h"
#include "main.h"

/* User interface */

// Graphics mode
int gr_mode = 0;

int windows_opened = 0;

// Position increment for new windows
#define WND_STEP_X 20;
#define WND_STEP_Y 20;

// Application instance handle
HINSTANCE 	g_hInstance;

// Window positions for new windows. After a new window is 
// created, wnd_x is incremented by WND_STEP_X and wnd_y
// is incremented by WND_STEP_Y.
int wnd_x = 0;
int wnd_y = 0;

/////////////////////////////////////////////////////////////////////////////

/* This is WindowProc callback common for all windows. This callback 
   ensures window painting and application termination when any of 
   windows is destroyed.
 */

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{

	switch (message) 
	{
		case WM_COMMAND:
			// Parse the menu selections:
   		        return DefWindowProc(hWnd, message, wParam, lParam);
			break;

		case WM_CREATE:
			break;
		case WM_PAINT: {
			
			HDC hDC;			
			PAINTSTRUCT ps;
			HBITMAP img;
			
			img = (HBITMAP)GetWindowLong(hWnd, 0); 
                        if (img == NULL) {
				return 0;
			}
			
			BITMAP bm;
			GetObject(img, sizeof(BITMAP), &bm);
									
			hDC = BeginPaint(hWnd, &ps);
   				HDC iDC = CreateCompatibleDC (hDC);
				SelectObject(iDC, img);
				BitBlt(hDC, 0, 0, bm.bmWidth, bm.bmHeight, iDC, 0, 0, SRCCOPY);  
				DeleteDC(iDC);
			EndPaint(hWnd, &ps);
		
			break;
		}	
		case WM_DESTROY:
			windows_opened--;
			if (windows_opened == 0) {
				PostQuitMessage(0);
			}		
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
   }
   return 0;
}

ATOM RegisterWindowClass(HINSTANCE hInstance, LPTSTR szWindowClass)
{
    WNDCLASS	wc;

    wc.style			= CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc		= (WNDPROC) WndProc;
    wc.cbClsExtra		= 0;
    wc.cbWndExtra		= 4; // single additional variable for bitmap pointer 
    wc.hInstance		= hInstance;
    wc.hIcon			= NULL;
    wc.hCursor			= 0;
    wc.hbrBackground		= (HBRUSH) GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName		= 0;
    wc.lpszClassName		= szWindowClass;

    return RegisterClass(&wc);
}

/////////////////////////////////////////////////////////////////////////////

extern "C" {

/* This function loads an RGB image from a file (*.bmp) */

#ifdef __CYGWIN__	

void load_rgb_image(char* filename, int* width, int* height, COLORREF* image) {

	leave_mmx_mode();

	if (width == NULL || height == NULL) {
		printf("Error: width and height parameter of the %s function must be a valid pointer\n", __FUNCTION__);
		exit(1);
	}	

	HANDLE img = LoadImage(NULL, filename, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
	if (img == NULL) {
		printf("Error: cannot load image from file %s\n", filename); 
		exit(1);
	}
	
	BITMAP bm;
	
	GetObject(img, sizeof(BITMAP), &bm);

	*width  = bm.bmWidth;
	*height = bm.bmHeight;
	
	printf("Load RGB image (%dx%d)\n", *width, *height);
						
        HDC hDC = CreateCompatibleDC(NULL);
	if (hDC == NULL) {
		DeleteObject(img);
		printf("Error: cannot creat compatible context (%d)\n", GetLastError());
		exit(1);
	}

	HGDIOBJ oimg = SelectObject(hDC, img);
	if (oimg == NULL || oimg == ((HGDIOBJ)GDI_ERROR)) {
		printf("Error: cannot select bitmap\n");
		DeleteObject(hDC); 
		DeleteObject(img); 
		exit(1);
	}
							
	if (image != NULL) {
		int x, y;
		for(y = 0; y < *height; y++) {
			for(x = 0; x < *width; x++) {
				image[y * (*width) + x] = GetPixel(hDC, x, y);
			}
		}
	}				 

	SelectObject(hDC, oimg);
	DeleteObject(hDC);
	DeleteObject(img);
	enter_mmx_mode();
}	

#endif

/////////////////////////////////////////////////////////////////////////////

/* This function loads an RGB image from a file and convert it to the
   grayscale image. Each 8-bit pixel represents 256-gray levels.
 */ 
  
void load_gray_image(char* filename, int* width, int* height, unsigned char* image) {

	leave_mmx_mode();

	if (width == NULL || height == NULL) {
		printf("Error: width and height parameter of the %s function must be a valid pointer\n", __FUNCTION__);
		exit(1);
	}	
	
        HDC hDC = CreateCompatibleDC(NULL);
	if (hDC == NULL) {
		printf("Error: cannot creat compatible context (%d)\n", GetLastError());
		exit(1);
	}
	
	HANDLE img = LoadImage(NULL, filename, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
	if (img == NULL) {
		printf("Error: cannot load image from file %s\n", filename); 
		exit(1);
	}
	
	BITMAP bm;
	
	GetObject(img, sizeof(BITMAP), &bm);

	*width = bm.bmWidth;
	*height = bm.bmHeight;
	
	printf("Load gray image (%dx%d)\n", *width, *height);
	
	HGDIOBJ oimg = SelectObject(hDC, img);
	if (oimg == NULL || oimg == ((HGDIOBJ)GDI_ERROR)) {
		printf("Error: cannot select bitmap\n"); 
		exit(1);
	}
	
								
	if (image != NULL) {
		int x, y;
		for(y = 0; y < *height; y++) {
			for(x = 0; x < *width; x++) {
				COLORREF pixel = GetPixel(hDC, x, y);
				int r = (pixel >> 16) & 0xFF;
				int g = (pixel >>  8) & 0xFF;
				int b = (pixel      ) & 0xFF;
				image[y * (*width) + x] =  (r + g + b) / 3;			}
		}
	}				 

	SelectObject(hDC, oimg);
	DeleteObject(hDC);
	DeleteObject(img);
	enter_mmx_mode();

}	

/* This function shows an RGB image in the 00bbggrr format.
 */

void show_rgb_image(int num, int width, int height, COLORREF* image) {

	leave_mmx_mode();

	if (!gr_mode) {
		printf("\nError: application is not in graphic mode\n\n");
		printHelp();
		exit(1);
	}	 

        CHAR title[100];
	
	wsprintf(title, TEXT("%.2d RGB Image"), num);
	
   	HWND wnd = CreateWindow(TEXT("Image Window"), title, 
		WS_SYSMENU, wnd_x, wnd_y, width, height, NULL, NULL, g_hInstance, NULL);
	if (wnd == NULL) { 
   		MessageBox(0, "Cannot create a new window", "Error", MB_OK);
		exit(1);
   	}

	wnd_x += WND_STEP_X;
	wnd_y += WND_STEP_Y;

   	RECT wr, cr;
   	GetWindowRect(wnd, &wr);
   	GetClientRect(wnd, &cr);

   	MoveWindow(wnd, wr.left, wr.top, 2 * width  - (cr.right  - cr.left), 2 * height - (cr.bottom - cr.top ), FALSE);
	
	HDC wDC = GetDC(wnd);
	if (wDC == NULL) {
		printf("Error: cannot get window context (%d)\n", GetLastError());
		exit(1);
	}
	
        HDC hDC = CreateCompatibleDC(wDC);
	if (hDC == NULL) {
		printf("Error: cannot creat compatible context (%d)\n", GetLastError());
		exit(1);
	}
	
	HBITMAP bm = CreateBitmap(width, height, 1, 32, image);
	
	HGDIOBJ oimg = SelectObject(hDC, bm);
	if (oimg == NULL || oimg == ((HGDIOBJ)GDI_ERROR)) {
		printf("Error: cannot select bitmap\n"); 
		exit(1);
	}

	printf("Show RGB image (%dx%d)\n", width, height);
	
	SetWindowLong(wnd, 0, (long int)bm);
	
	ShowWindow(wnd, true);

	windows_opened++;

	SelectObject(hDC, oimg);
	DeleteObject(hDC);
	enter_mmx_mode();	
}

/* This function shows a grayscale image. Each 8-bit pixel represents
   256 gray levels */ 

void show_gray_image(int num, int width, int height, unsigned char* image) {

	leave_mmx_mode();

	if (!gr_mode) {
		printf("\nError: application is not in graphic mode\n\n");
		printHelp();
		exit(1);
	}	 

        CHAR title[100];
	
	wsprintf(title, TEXT("%.2d RGB Image"), num);

	printf("Show gray image (%dx%d)\n", width, height);

   	HWND wnd = CreateWindow(TEXT("Image Window"), title, 
		WS_SYSMENU, wnd_x, wnd_y, width, height, NULL, NULL, g_hInstance, NULL);
	if (wnd == NULL) { 
   		MessageBox(0, "Cannot create a new window", "Error", MB_OK);
		exit(1);
   	}

	wnd_x += WND_STEP_X;
	wnd_y += WND_STEP_Y;

   	RECT wr, cr;
   	GetWindowRect(wnd, &wr);
   	GetClientRect(wnd, &cr);

   	MoveWindow(wnd, wr.left, wr.top, 2 * width  - (cr.right  - cr.left), 2 * height - (cr.bottom - cr.top ), FALSE);
		 
	HDC wDC = GetDC(wnd);
	if (wDC == NULL) {
		printf("Error: cannot get window context (%d)\n", GetLastError());
		exit(1);
	}
	
	HDC hDC = CreateCompatibleDC(wDC);
	if (hDC == NULL) {
		DeleteObject(wDC);
		DeleteObject(wnd);
		printf("Error: cannot careate compatible context (%d)\n", GetLastError());
		exit(1);
	}
			 	 					
	HBITMAP bm = CreateBitmap(width, height, 1, 32, NULL);

	HGDIOBJ oimg = SelectObject(hDC, bm);

	if (oimg == NULL || oimg == ((HGDIOBJ)GDI_ERROR)) {
		DeleteObject(bm);
		DeleteObject(wDC);
		DeleteObject(hDC);
		DeleteObject(wnd);
		printf("Error: cannot select bitmap\n"); 
		exit(1);
	}


	for(int y = 0; y < height; y++) {
		for(int x = 0; x < width; x++) {
			unsigned char v = image[y * width + x];
			COLORREF pixel = v << 16 | v << 8 | v;
			SetPixel(hDC, x, y, pixel); 
		}
	}	 	
	
	SelectObject(hDC, oimg);
	
	SetWindowLong(wnd, 0, (long int)bm);

	ShowWindow(wnd, true);
	
	windows_opened++;
		
	DeleteObject(wDC);
	DeleteObject(hDC);
	enter_mmx_mode();
	
}

void main_program_loop() {
}

}


/////////////////////////////////////////////////////////////////////////////

/* This function parses command line options and stores them
   in char* argv[] structure. Function returns the number of 
   arguments. If argv is NULL, function parses the command line,
   and returns the number of arguments only.
    
*/    

#ifdef __CYGWIN__

int ParseCmdLine(char* CmdLine, char** argv = NULL) {
	int argc = 0;
	char* s = CmdLine;
	while (true) {
		// remove spaces		
		while ((*s) != '\0' && (*s)==' ') s++;

		if ((*s)== '\0') { 
			break;
		}

		if (argv != NULL) {
			argv[argc] = s;
		}		
		// read string
		while ((*s) != '\0' && (*s)!=' ') s++;
		argc++;
		if ((*s)== '\0') { 
			break;
		} else {
		 	if (argv != NULL) (*s++) = '\0';
		}			
	}  
	return argc;
}

#endif

/////////////////////////////////////////////////////////////////////////////

/* The main Window function */
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow) {

        // Store application instance handle 
   	g_hInstance = hInstance;
   	
	// Create window class for   
   	RegisterWindowClass(g_hInstance, "Image Window");

   	char* cmdLine = strdup(GetCommandLine());
      
   	int argc = ParseCmdLine(cmdLine);
      
   	char** argv = new char*[argc]; 
   
   	ParseCmdLine(cmdLine, argv);

	int retval = main_body(argc, argv);
	
	if (retval) {
		return retval;
	}
	         
	/* The main graphic loop */
	if (gr_mode) { 
		MSG msg;
  		while (GetMessage(&msg, NULL, 0, 0)) {
     			TranslateMessage(&msg);
     			DispatchMessage(&msg);      
  		}
		
	}	
 
  	return 0;
}

/////////////////////////////////////////////////////////////////////////////
